/*
# This file is part of libkd.
# Licensed under a 3-clause BSD style license - see LICENSE
*/

typedef float ttype;

#define BIGTTYPE float
#define BIGTTYPE_MAX KDT_INFTY_FLOAT
typedef float bigttype;

#define TTYPE_INTEGER 0

#define TTYPE_MIN -KDT_INFTY_FLOAT
#define TTYPE_MAX  KDT_INFTY_FLOAT
#define TTYPE_SQRT_MAX  1e19

#define TTYPE f
#define TTYPE_M f

