/*
# This file is part of libkd.
# Licensed under a 3-clause BSD style license - see LICENSE
*/

typedef double etype;

#define ETYPE_INTEGER 0

#define ETYPE_MAX  KDT_INFTY_DOUBLE
#define ETYPE_MIN -KDT_INFTY_DOUBLE

#define ETYPE d
#define ETYPE_M d

