/*
# This file is part of libkd.
# Licensed under a 3-clause BSD style license - see LICENSE
*/

typedef u32 dtype;

#define DTYPE_INTEGER 1
#define DTYPE_DOUBLE  0

#define DTYPE_MAX  UINT32_MAX
#define DTYPE_MIN  0

#define DTYPE u
#define DTYPE_M u

#define DTYPE_KDT_DATA  KDT_DATA_U32
