/*
# This file is part of libkd.
# Licensed under a 3-clause BSD style license - see LICENSE
*/

typedef float dtype;

#define DTYPE_INTEGER 0
#define DTYPE_DOUBLE  0

#define DTYPE_MAX  KDT_INFTY_FLOAT
#define DTYPE_MIN -KDT_INFTY_FLOAT

#define DTYPE f
#define DTYPE_M f

#define DTYPE_KDT_DATA  KDT_DATA_FLOAT
